@@ -26,24 +26,21 @@
 #include <config.h>
 #endif
 
-#include <dbus/dbus-glib.h>
+#include <stdlib.h>
 
 #include <glib/gi18n.h>
-
 #include <gtk/gtk.h>
 
+#include <dbus/dbus-glib.h>
+
+#include <bluetooth-client.h>
+#include <bluetooth-agent.h>
+
 #include "notify.h"
 #include "agent.h"
 
-#define PASSKEY_AGENT_PATH	"/org/bluez/passkey"
-#define AUTH_AGENT_PATH		"/org/bluez/auth"
-
-static DBusGConnection *connection;
-
-static int volatile registered_passkey	= 0;
-static int volatile registered_auth	= 0;
-
-static gboolean auto_authorize = FALSE;
+static BluetoothClient *client;
+static GtkTreeModel *adapter_model;
 
 typedef enum {
 	AGENT_ERROR_REJECT
@@ -83,9 +80,9 @@ static GList *input_list = NULL;
 
 struct input_data {
 	char *path;
-	char *address;
-	char *service;
 	char *uuid;
+	gboolean numeric;
+	DBusGProxy *device;
 	DBusGMethodInvocation *context;
 	GtkWidget *dialog;
 	GtkWidget *button;
@@ -97,7 +94,7 @@ static gint input_compare(gconstpointer a, gconstpointer b)
 	struct input_data *a_data = (struct input_data *) a;
 	struct input_data *b_data = (struct input_data *) b;
 
-	return g_ascii_strcasecmp(a_data->address, b_data->address);
+	return g_ascii_strcasecmp(a_data->path, b_data->path);
 }
 
 static void input_free(struct input_data *input)
@@ -106,9 +103,10 @@ static void input_free(struct input_data *input)
 
 	input_list = g_list_remove(input_list, input);
 
+	if (input->device != NULL)
+		g_object_unref(input->device);
+
 	g_free(input->uuid);
-	g_free(input->service);
-	g_free(input->address);
 	g_free(input->path);
 	g_free(input);
 
@@ -122,9 +120,14 @@ static void passkey_callback(GtkWidget *dialog,
 	struct input_data *input = user_data;
 
 	if (response == GTK_RESPONSE_ACCEPT) {
-		const char *passkey;
-		passkey = gtk_entry_get_text(GTK_ENTRY(input->entry));
-		dbus_g_method_return(input->context, passkey);
+		const char *text;
+		text = gtk_entry_get_text(GTK_ENTRY(input->entry));
+
+		if (input->numeric == TRUE) {
+			guint passkey = atoi(text);
+			dbus_g_method_return(input->context, passkey);
+		} else
+			dbus_g_method_return(input->context, text);
 	} else {
 		GError *error;
 		error = g_error_new(AGENT_ERROR, AGENT_ERROR_REJECT,
@@ -135,7 +138,6 @@ static void passkey_callback(GtkWidget *dialog,
 	input_free(input);
 }
 
-#if 0
 static void confirm_callback(GtkWidget *dialog,
 				gint response, gpointer user_data)
 {
@@ -151,23 +153,27 @@ static void confirm_callback(GtkWidget *dialog,
 
 	input_free(input);
 }
-#endif
 
 static void set_trusted(struct input_data *input)
 {
-	DBusGProxy *object;
+	GValue value = { 0 };
 	gboolean active;
 
+	if (input->device == NULL)
+		return;
+
 	active = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(input->button));
 	if (active == FALSE)
 		return;
 
-	object = dbus_g_proxy_new_for_name(connection, "org.bluez",
-					input->path, "org.bluez.Adapter");
+	g_value_init(&value, G_TYPE_BOOLEAN);
+	g_value_set_boolean(&value, TRUE);
 
-	dbus_g_proxy_call(object, "SetTrusted", NULL,
-				G_TYPE_STRING, input->address, G_TYPE_INVALID,
-								G_TYPE_INVALID);
+	dbus_g_proxy_call(input->device, "SetProperty", NULL,
+			G_TYPE_STRING, "Trusted",
+			G_TYPE_VALUE, &value, G_TYPE_INVALID, G_TYPE_INVALID);
+
+	g_value_unset(&value);
 }
 
 static void auth_callback(GtkWidget *dialog,
@@ -188,6 +194,23 @@ static void auth_callback(GtkWidget *dialog,
 	input_free(input);
 }
 
+static void insert_callback(GtkEditable *editable, const gchar *text,
+			gint length, gint *position, gpointer user_data)
+{
+	struct input_data *input = user_data;
+	gint i;
+
+	if (input->numeric == FALSE)
+		return;
+
+	for (i = 0; i < length; i++) {
+		if (g_ascii_isdigit(text[i]) == FALSE) {
+			g_signal_stop_emission_by_name(editable,
+							"insert-text");
+		}
+	}
+}
+
 static void changed_callback(GtkWidget *editable, gpointer user_data)
 {
 	struct input_data *input = user_data;
@@ -208,8 +231,9 @@ static void toggled_callback(GtkWidget *button, gpointer user_data)
 	gtk_entry_set_visibility(GTK_ENTRY(input->entry), mode);
 }
 
-static void passkey_dialog(const char *path, const char *address,
-			const gchar *device, DBusGMethodInvocation *context)
+static void passkey_dialog(DBusGProxy *adapter, DBusGProxy *device,
+		const char *address, const char *name, gboolean numeric,
+						DBusGMethodInvocation *context)
 {
 	GtkWidget *dialog;
 	GtkWidget *button;
@@ -225,115 +249,86 @@ static void passkey_dialog(const char *path, const char *address,
 	if (!input)
 		return;
 
-	input->path = g_strdup(path);
-	input->address = g_strdup(address);
-
+	input->path = g_strdup(dbus_g_proxy_get_path(adapter));
+	input->numeric = numeric;
 	input->context = context;
+	input->device = g_object_ref(device);
 
 	dialog = gtk_dialog_new();
-
 	gtk_window_set_title(GTK_WINDOW(dialog), _("Authentication request"));
-
 	gtk_window_set_resizable(GTK_WINDOW(dialog), FALSE);
-
 	gtk_window_set_position(GTK_WINDOW(dialog), GTK_WIN_POS_CENTER);
-
 	gtk_window_set_keep_above(GTK_WINDOW(dialog), TRUE);
-
 	gtk_window_set_urgency_hint(GTK_WINDOW(dialog), TRUE);
-
 	gtk_dialog_set_has_separator(GTK_DIALOG(dialog), FALSE);
-
 	input->dialog = dialog;
 
 	button = gtk_dialog_add_button(GTK_DIALOG(dialog),
 				GTK_STOCK_CANCEL, GTK_RESPONSE_REJECT);
-
 	button = gtk_dialog_add_button(GTK_DIALOG(dialog),
 					GTK_STOCK_OK, GTK_RESPONSE_ACCEPT);
-
 	gtk_widget_grab_default(button);
-
 	gtk_widget_set_sensitive(button, FALSE);
-
 	input->button = button;
 
 	table = gtk_table_new(5, 2, FALSE);
-
 	gtk_table_set_row_spacings(GTK_TABLE(table), 4);
 	gtk_table_set_col_spacings(GTK_TABLE(table), 20);
-
 	gtk_container_set_border_width(GTK_CONTAINER(table), 12);
-
 	gtk_container_add(GTK_CONTAINER(GTK_DIALOG(dialog)->vbox), table);
-
 	image = gtk_image_new_from_icon_name(GTK_STOCK_DIALOG_AUTHENTICATION,
 							GTK_ICON_SIZE_DIALOG);
-
 	gtk_misc_set_alignment(GTK_MISC(image), 0.0, 0.0);
-
 	gtk_table_attach(GTK_TABLE(table), image, 0, 1, 0, 5,
 						GTK_SHRINK, GTK_FILL, 0, 0);
-
 	vbox = gtk_vbox_new(FALSE, 6);
 
 	label = gtk_label_new(_("Pairing request for device:"));
-
 	gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.0);
-
 	gtk_container_add(GTK_CONTAINER(vbox), label);
-
 	gtk_table_attach(GTK_TABLE(table), vbox, 1, 2, 0, 1,
 				GTK_EXPAND | GTK_FILL, GTK_SHRINK, 0, 0);
 
 	label = gtk_label_new(NULL);
-
-	markup = g_strdup_printf("<b>%s</b>", device);
+	markup = g_strdup_printf("<b>%s</b>", name);
 	gtk_label_set_markup(GTK_LABEL(label), markup);
 	g_free(markup);
-
 	gtk_label_set_line_wrap(GTK_LABEL(label), TRUE);
-
 	gtk_label_set_selectable(GTK_LABEL(label), TRUE);
-
 	gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.0);
-
 	gtk_widget_set_size_request(GTK_WIDGET(label), 280, -1);
-
 	gtk_container_add(GTK_CONTAINER(vbox), label);
-
 	vbox = gtk_vbox_new(FALSE, 6);
 
 	label = gtk_label_new(_("Enter passkey for authentication:"));
-
 	gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.0);
-
 	gtk_container_add(GTK_CONTAINER(vbox), label);
-
 	gtk_table_attach(GTK_TABLE(table), vbox, 1, 2, 2, 3,
 				GTK_EXPAND | GTK_FILL, GTK_SHRINK, 0, 0);
 
 	entry = gtk_entry_new();
-
-	gtk_entry_set_max_length(GTK_ENTRY(entry), 16);
-
+	if (numeric == TRUE) {
+		gtk_entry_set_max_length(GTK_ENTRY(entry), 6);
+		gtk_entry_set_width_chars(GTK_ENTRY(entry), 6);
+		g_signal_connect(G_OBJECT(entry), "insert-text",
+					G_CALLBACK(insert_callback), input);
+	} else {
+		gtk_entry_set_max_length(GTK_ENTRY(entry), 16);
+		gtk_entry_set_width_chars(GTK_ENTRY(entry), 16);
+		gtk_entry_set_visibility(GTK_ENTRY(entry), FALSE);
+	}
 	gtk_entry_set_activates_default(GTK_ENTRY(entry), TRUE);
-
-	gtk_entry_set_visibility(GTK_ENTRY(entry), FALSE);
-
 	input->entry = entry;
-
 	g_signal_connect(G_OBJECT(entry), "changed",
 				G_CALLBACK(changed_callback), input);
-
 	gtk_container_add(GTK_CONTAINER(vbox), entry);
 
-	button = gtk_check_button_new_with_label(_("Show input"));
-
-	g_signal_connect(G_OBJECT(button), "toggled",
-				G_CALLBACK(toggled_callback), input);
-
-	gtk_container_add(GTK_CONTAINER(vbox), button);
+	if (numeric == FALSE) {
+		button = gtk_check_button_new_with_label(_("Show input"));
+		g_signal_connect(G_OBJECT(button), "toggled",
+					G_CALLBACK(toggled_callback), input);
+		gtk_container_add(GTK_CONTAINER(vbox), button);
+	}
 
 	input_list = g_list_append(input_list, input);
 
@@ -343,10 +338,15 @@ static void passkey_dialog(const char *path, const char *address,
 	enable_blinking();
 }
 
-#if 0
-static void confirm_dialog(const char *path, const char *address,
-				const char *value, const gchar *device,
-					DBusGMethodInvocation *context)
+static void display_dialog(DBusGProxy *adapter, DBusGProxy *device,
+		const char *address, const char *name, const char *value,
+				guint entered, DBusGMethodInvocation *context)
+{
+}
+
+static void confirm_dialog(DBusGProxy *adapter, DBusGProxy *device,
+		const char *address, const char *name, const char *value,
+						DBusGMethodInvocation *context)
 {
 	GtkWidget *dialog;
 	GtkWidget *button;
@@ -361,97 +361,65 @@ static void confirm_dialog(const char *path, const char *address,
 	if (!input)
 		return;
 
-	input->path = g_strdup(path);
-	input->address = g_strdup(address);
-
+	input->path = g_strdup(dbus_g_proxy_get_path(adapter));
+	input->device = g_object_ref(device);
 	input->context = context;
 
 	dialog = gtk_dialog_new();
-
 	gtk_window_set_title(GTK_WINDOW(dialog), _("Confirmation request"));
-
 	gtk_window_set_resizable(GTK_WINDOW(dialog), FALSE);
-
 	gtk_window_set_position(GTK_WINDOW(dialog), GTK_WIN_POS_CENTER);
-
 	gtk_window_set_keep_above(GTK_WINDOW(dialog), TRUE);
-
 	gtk_window_set_urgency_hint(GTK_WINDOW(dialog), TRUE);
-
 	gtk_dialog_set_has_separator(GTK_DIALOG(dialog), FALSE);
-
 	input->dialog = dialog;
 
 	button = gtk_dialog_add_button(GTK_DIALOG(dialog),
 					GTK_STOCK_NO, GTK_RESPONSE_NO);
- 
 	button = gtk_dialog_add_button(GTK_DIALOG(dialog),
 					GTK_STOCK_YES, GTK_RESPONSE_YES);
 
 	table = gtk_table_new(5, 2, FALSE);
-
 	gtk_table_set_row_spacings(GTK_TABLE(table), 4);
 	gtk_table_set_col_spacings(GTK_TABLE(table), 20);
-
 	gtk_container_set_border_width(GTK_CONTAINER(table), 12);
-
 	gtk_container_add(GTK_CONTAINER(GTK_DIALOG(dialog)->vbox), table);
 
 	image = gtk_image_new_from_icon_name(GTK_STOCK_DIALOG_AUTHENTICATION,
 							GTK_ICON_SIZE_DIALOG);
-
 	gtk_misc_set_alignment(GTK_MISC(image), 0.0, 0.0);
-
 	gtk_table_attach(GTK_TABLE(table), image, 0, 1, 0, 5,
 						GTK_SHRINK, GTK_FILL, 0, 0);
 
 	vbox = gtk_vbox_new(FALSE, 6);
 	label = gtk_label_new(_("Pairing request for device:"));
-
 	gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.0);
-
 	gtk_container_add(GTK_CONTAINER(vbox), label);
-
 	gtk_table_attach(GTK_TABLE(table), vbox, 1, 2, 0, 1,
 				GTK_EXPAND | GTK_FILL, GTK_SHRINK, 0, 0);
 
 	label = gtk_label_new(NULL);
-
-	markup = g_strdup_printf("<b>%s</b>", device);
+	markup = g_strdup_printf("<b>%s</b>", name);
 	gtk_label_set_markup(GTK_LABEL(label), markup);
 	g_free(markup);
-
 	gtk_label_set_line_wrap(GTK_LABEL(label), TRUE);
-
 	gtk_label_set_selectable(GTK_LABEL(label), TRUE);
-
 	gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.0);
-
 	gtk_widget_set_size_request(GTK_WIDGET(label), 280, -1);
-
 	gtk_container_add(GTK_CONTAINER(vbox), label);
 
 	vbox = gtk_vbox_new(FALSE, 6);
-
 	label = gtk_label_new(_("Confirm value for authentication:"));
-
 	gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.0);
-
 	gtk_container_add(GTK_CONTAINER(vbox), label);
-
 	gtk_table_attach(GTK_TABLE(table), vbox, 1, 2, 2, 3,
 				GTK_EXPAND | GTK_FILL, GTK_SHRINK, 0, 0);
 
 	label = gtk_label_new(NULL);
-
 	markup = g_strdup_printf("<b>%s</b>\n", value);
-
 	gtk_label_set_markup(GTK_LABEL(label), markup);
-
 	g_free(markup);
-
 	gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.0);
-
 	gtk_container_add(GTK_CONTAINER(vbox), label);
 
 	input_list = g_list_append(input_list, input);
@@ -461,11 +429,10 @@ static void confirm_dialog(const char *path, const char *address,
 
 	enable_blinking();
 }
-#endif
 
-static void auth_dialog(const char *path, const char *address,
-		const char *service, const char *uuid, const gchar *device,
-		const gchar *profile, DBusGMethodInvocation *context)
+static void auth_dialog(DBusGProxy *adapter, DBusGProxy *device,
+		const char *address, const char *name, const char *uuid,
+						DBusGMethodInvocation *context)
 {
 	GtkWidget *dialog;
 	GtkWidget *button;
@@ -480,102 +447,70 @@ static void auth_dialog(const char *path, const char *address,
 	if (!input)
 		return;
 
-	input->path = g_strdup(path);
-	input->address = g_strdup(address);
-	input->service = g_strdup(service);
+	input->path = g_strdup(dbus_g_proxy_get_path(adapter));
 	input->uuid = g_strdup(uuid);
-
+	input->device = g_object_ref(device);
 	input->context = context;
 
 	dialog = gtk_dialog_new();
-
 	gtk_window_set_title(GTK_WINDOW(dialog), _("Authorization request"));
-
 	gtk_window_set_resizable(GTK_WINDOW(dialog), FALSE);
-
 	gtk_window_set_position(GTK_WINDOW(dialog), GTK_WIN_POS_CENTER);
-
 	gtk_window_set_keep_above(GTK_WINDOW(dialog), TRUE);
-
 	gtk_window_set_urgency_hint(GTK_WINDOW(dialog), TRUE);
-
 	gtk_dialog_set_has_separator(GTK_DIALOG(dialog), FALSE);
-
 	input->dialog = dialog;
 
 	button = gtk_dialog_add_button(GTK_DIALOG(dialog),
 					GTK_STOCK_NO, GTK_RESPONSE_NO);
- 
 	button = gtk_dialog_add_button(GTK_DIALOG(dialog),
 					GTK_STOCK_YES, GTK_RESPONSE_YES);
 
 	table = gtk_table_new(5, 2, FALSE);
-
 	gtk_table_set_row_spacings(GTK_TABLE(table), 4);
 	gtk_table_set_col_spacings(GTK_TABLE(table), 20);
-
 	gtk_container_set_border_width(GTK_CONTAINER(table), 12);
-
 	gtk_container_add(GTK_CONTAINER(GTK_DIALOG(dialog)->vbox), table);
 
 	image = gtk_image_new_from_icon_name(GTK_STOCK_DIALOG_AUTHENTICATION,
 							GTK_ICON_SIZE_DIALOG);
-
 	gtk_misc_set_alignment(GTK_MISC(image), 0.0, 0.0);
-
 	gtk_table_attach(GTK_TABLE(table), image, 0, 1, 0, 5,
 						GTK_SHRINK, GTK_FILL, 0, 0);
 
 	vbox = gtk_vbox_new(FALSE, 6);
-
 	label = gtk_label_new(_("Authorization request for device:"));
-
 	gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.0);
-
 	gtk_table_attach(GTK_TABLE(table), label, 1, 2, 0, 1,
 				GTK_EXPAND | GTK_FILL, GTK_SHRINK, 0, 0);
 
 	label = gtk_label_new(NULL);
-
-	markup = g_strdup_printf("<b>%s</b>", device);
+	markup = g_strdup_printf("<b>%s</b>", name);
 	gtk_label_set_markup(GTK_LABEL(label), markup);
 	g_free(markup);
-
 	gtk_label_set_line_wrap(GTK_LABEL(label), TRUE);
-
 	gtk_label_set_selectable(GTK_LABEL(label), TRUE);
-
 	gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.0);
-
 	gtk_widget_set_size_request(GTK_WIDGET(label), 280, -1);
-
 	gtk_container_add(GTK_CONTAINER(vbox), label);
-
 	gtk_table_attach(GTK_TABLE(table), vbox, 1, 2, 2, 3,
 				GTK_EXPAND | GTK_FILL, GTK_SHRINK, 0, 0);
 
 	label = gtk_label_new(NULL);
-
 	/* translators: Whether to grant access to a particular service
 	 * to the device mentioned */
-	markup = g_strdup_printf("<i>%s</i>", profile);
+	markup = g_strdup_printf("<i>%s</i>", uuid);
 	text = g_strdup_printf(_("Grant access to %s?"), markup);
 	g_free(markup);
-
 	markup = g_strdup_printf("%s\n", text);
 	gtk_label_set_markup(GTK_LABEL(label), markup);
 	g_free(markup);
-
 	g_free(text);
-
 	gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.0);
-
 	gtk_container_add(GTK_CONTAINER(vbox), label);
 
 	button = gtk_check_button_new_with_label(_("Always grant access"));
-
 	input->button = button;
-
 	gtk_container_add(GTK_CONTAINER(vbox), button);
 
 	input_list = g_list_append(input_list, input);
@@ -595,53 +530,6 @@ static void show_dialog(gpointer data, gpointer user_data)
 	gtk_window_present(GTK_WINDOW(input->dialog));
 }
 
-typedef struct {
-	GObject parent;
-} PasskeyAgent;
-
-typedef struct {
-	GObjectClass parent;
-} PasskeyAgentClass;
-
-static GObjectClass *passkey_agent_parent;
-
-G_DEFINE_TYPE(PasskeyAgent, passkey_agent, G_TYPE_OBJECT)
-
-#define PASSKEY_AGENT_OBJECT_TYPE (passkey_agent_get_type())
-
-#define PASSKEY_AGENT(obj) (G_TYPE_CHECK_INSTANCE_CAST((obj), \
-				PASSKEY_AGENT_OBJECT_TYPE, PasskeyAgent))
-
-static void passkey_agent_finalize(GObject *obj)
-{
-	passkey_agent_parent->finalize(obj);
-}
-
-static void passkey_agent_init(PasskeyAgent *obj)
-{
-}
-
-static void passkey_agent_class_init(PasskeyAgentClass *klass)
-{
-	GObjectClass *gobject_class;
-
-	passkey_agent_parent = g_type_class_peek_parent(klass);
-
-	gobject_class = G_OBJECT_CLASS(klass);
-	gobject_class->finalize = passkey_agent_finalize;
-}
-
-static PasskeyAgent *passkey_agent_new(const char *path)
-{
-	PasskeyAgent *agent;
-
-	agent = g_object_new(PASSKEY_AGENT_OBJECT_TYPE, NULL);
-
-	dbus_g_connection_register_g_object(connection, path, G_OBJECT(agent));
-
-	return agent;
-}
-
 static void notification_closed(GObject *object, gpointer user_data)
 {
 	g_list_foreach(input_list, show_dialog, NULL);
@@ -649,40 +537,104 @@ static void notification_closed(GObject *object, gpointer user_data)
 	disable_blinking();
 }
 
-static gboolean passkey_agent_request(PasskeyAgent *agent,
-				const char *path, const char *address,
-					DBusGMethodInvocation *context)
+#ifndef DBUS_TYPE_G_DICTIONARY
+#define DBUS_TYPE_G_DICTIONARY \
+	(dbus_g_type_get_map("GHashTable", G_TYPE_STRING, G_TYPE_VALUE))
+#endif
+
+static gboolean device_get_properties(DBusGProxy *proxy,
+					GHashTable **hash, GError **error)
 {
-	DBusGProxy *object;
-	const char *adapter = NULL, *name = NULL;
-	gchar *device, *line;
+	return dbus_g_proxy_call(proxy, "GetProperties", error,
+		G_TYPE_INVALID, DBUS_TYPE_G_DICTIONARY, hash, G_TYPE_INVALID);
+}
 
-	object = dbus_g_proxy_new_for_name(connection, "org.bluez",
-						path, "org.bluez.Adapter");
+static gboolean pincode_request(DBusGMethodInvocation *context,
+					DBusGProxy *device, gpointer user_data)
+{
+	DBusGProxy *adapter = user_data;
+	GHashTable *hash = NULL;
+	GValue *value;
+	const gchar *address, *alias;
+	gchar *name, *line;
 
-	dbus_g_proxy_call(object, "GetName", NULL, G_TYPE_INVALID,
-				G_TYPE_STRING, &adapter, G_TYPE_INVALID);
+	device_get_properties(device, &hash, NULL);
 
-	dbus_g_proxy_call(object, "GetRemoteName", NULL,
-				G_TYPE_STRING, address, G_TYPE_INVALID,
-				G_TYPE_STRING, &name, G_TYPE_INVALID);
+	if (hash != NULL) {
+		value = g_hash_table_lookup(hash, "Address");
+		address = value ? g_value_get_string(value) : NULL;
 
-	if (name) {
-		if (g_strrstr(name, address))
-			device = g_strdup(name);
+		value = g_hash_table_lookup(hash, "Name");
+		alias = value ? g_value_get_string(value) : NULL;
+	} else {
+		address = NULL;
+		alias = NULL;
+	}
+
+	if (alias) {
+		if (g_strrstr(alias, address))
+			name = g_strdup(alias);
 		else
-			device = g_strdup_printf("%s (%s)", name, address);
+			name = g_strdup_printf("%s (%s)", alias, address);
 	} else
-		device = g_strdup(address);
+		name = g_strdup(address);
 
-	passkey_dialog(path, address, device, context);
+	passkey_dialog(adapter, device, address, name, FALSE, context);
 
 	/* translators: this is a popup telling you a particular device
 	 * has asked for pairing */
-	line = g_strdup_printf(_("Pairing request for '%s'"), device);
-	g_free(device);
+	line = g_strdup_printf(_("Pairing request for %s"), name);
 
-	show_notification(adapter ? adapter : _("Bluetooth device"),
+	g_free(name);
+
+	show_notification(_("Bluetooth device"),
+					line, _("Enter PIN code"), 0,
+					G_CALLBACK(notification_closed));
+
+	g_free(line);
+
+	return TRUE;
+}
+
+static gboolean passkey_request(DBusGMethodInvocation *context,
+					DBusGProxy *device, gpointer user_data)
+{
+	DBusGProxy *adapter = user_data;
+	GHashTable *hash = NULL;
+	GValue *value;
+	const gchar *address, *alias;
+	gchar *name, *line;
+
+	device_get_properties(device, &hash, NULL);
+
+	if (hash != NULL) {
+		value = g_hash_table_lookup(hash, "Address");
+		address = value ? g_value_get_string(value) : NULL;
+
+		value = g_hash_table_lookup(hash, "Name");
+		alias = value ? g_value_get_string(value) : NULL;
+	} else {
+		address = NULL;
+		alias = NULL;
+	}
+
+	if (alias) {
+		if (g_strrstr(alias, address))
+			name = g_strdup(alias);
+		else
+			name = g_strdup_printf("%s (%s)", alias, address);
+	} else
+		name = g_strdup(address);
+
+	passkey_dialog(adapter, device, address, name, TRUE, context);
+
+	/* translators: this is a popup telling you a particular device
+	 * has asked for pairing */
+	line = g_strdup_printf(_("Pairing request for %s"), name);
+
+	g_free(name);
+
+	show_notification(_("Bluetooth device"),
 					line, _("Enter passkey"), 0,
 					G_CALLBACK(notification_closed));
 
@@ -691,184 +643,142 @@ static gboolean passkey_agent_request(PasskeyAgent *agent,
 	return TRUE;
 }
 
-#if 0
-static gboolean passkey_agent_confirm(PasskeyAgent *agent,
-			const char *path, const char *address,
-			const char *value, DBusGMethodInvocation *context)
+static gboolean display_request(DBusGMethodInvocation *context,
+				DBusGProxy *device, guint passkey,
+					guint entered, gpointer user_data)
 {
-	DBusGProxy *object;
-	const char *adapter = NULL, *name = NULL;
-	gchar *device, *line;
+	DBusGProxy *adapter = user_data;
+	GHashTable *hash = NULL;
+	GValue *value;
+	const gchar *address, *alias;
+	gchar *name, *line, *text;
 
-	object = dbus_g_proxy_new_for_name(connection, "org.bluez",
-						path, "org.bluez.Adapter");
+	device_get_properties(device, &hash, NULL);
 
-	dbus_g_proxy_call(object, "GetName", NULL, G_TYPE_INVALID,
-				G_TYPE_STRING, &adapter, G_TYPE_INVALID);
+	if (hash != NULL) {
+		value = g_hash_table_lookup(hash, "Address");
+		address = value ? g_value_get_string(value) : NULL;
 
-	dbus_g_proxy_call(object, "GetRemoteName", NULL,
-				G_TYPE_STRING, address, G_TYPE_INVALID,
-				G_TYPE_STRING, &name, G_TYPE_INVALID);
-
-	if (name) {
-		if (g_strrstr(name, address))
-			device = g_strdup(name);
-		else
-			device = g_strdup_printf("%s (%s)", name, address);
-	} else
-		device = g_strdup(address);
-
-	confirm_dialog(path, address, value, device, context);
-
-	line = g_strdup_printf(_("Pairing request for '%s'"), device);
-	g_free(device);
-
-	show_notification(adapter ? adapter : _("Bluetooth device"),
-					line, _("Confirm pairing"), 0,
-					G_CALLBACK(notification_closed));
-
-	g_free (line);
-
-	return TRUE;
-}
-#endif
-
-static gboolean passkey_agent_cancel(PasskeyAgent *agent,
-			const char *path, const char *address, GError **error)
-{
-	GList *list;
-	GError *result;
-	struct input_data *input;
-
-	input = g_try_malloc0(sizeof(*input));
-	if (!input)
-		return FALSE;
-
-	input->path = g_strdup(path);
-	input->address = g_strdup(address);
-
-	list = g_list_find_custom(input_list, input, input_compare);
-
-	g_free(input->address);
-	g_free(input->path);
-	g_free(input);
-
-	if (!list || !list->data)
-		return FALSE;
-
-	input = list->data;
-
-	close_notification();
-
-	result = g_error_new(AGENT_ERROR, AGENT_ERROR_REJECT,
-						"Agent callback canceled");
-
-	dbus_g_method_return_error(input->context, result);
-
-	input_free(input);
-
-	return TRUE;
-}
-
-static gboolean passkey_agent_release(PasskeyAgent *agent, GError **error)
-{
-	registered_passkey = 0;
-
-	return TRUE;
-}
-
-#include "passkey-agent-glue.h"
-
-typedef struct {
-	GObject parent;
-} AuthAgent;
-
-typedef struct {
-	GObjectClass parent;
-} AuthAgentClass;
-
-static GObjectClass *auth_agent_parent;
-
-G_DEFINE_TYPE(AuthAgent, auth_agent, G_TYPE_OBJECT)
-
-#define AUTH_AGENT_OBJECT_TYPE (auth_agent_get_type())
-
-#define AUTH_AGENT(obj) (G_TYPE_CHECK_INSTANCE_CAST((obj), \
-				AUTH_AGENT_OBJECT_TYPE, AuthAgent))
-
-static void auth_agent_finalize(GObject *obj)
-{
-	auth_agent_parent->finalize(obj);
-}
-
-static void auth_agent_init(AuthAgent *obj)
-{
-}
-
-static void auth_agent_class_init(AuthAgentClass *klass)
-{
-	GObjectClass *gobject_class;
-
-	auth_agent_parent = g_type_class_peek_parent(klass);
-
-	gobject_class = G_OBJECT_CLASS(klass);
-	gobject_class->finalize = auth_agent_finalize;
-}
-
-static AuthAgent *auth_agent_new(const char *path)
-{
-	AuthAgent *agent;
-
-	agent = g_object_new(AUTH_AGENT_OBJECT_TYPE, NULL);
-
-	dbus_g_connection_register_g_object(connection, path, G_OBJECT(agent));
-
-	return agent;
-}
-
-static gboolean auth_agent_authorize(PasskeyAgent *agent,
-		const char *path, const char *address, const char *service,
-			const char *uuid, DBusGMethodInvocation *context)
-{
-	DBusGProxy *object;
-	const char *adapter = NULL, *name = NULL;
-	gchar *device, *profile, *line;
-
-	if (auto_authorize == TRUE) {
-		dbus_g_method_return(context);
-		return TRUE;
+		value = g_hash_table_lookup(hash, "Name");
+		alias = value ? g_value_get_string(value) : NULL;
+	} else {
+		address = NULL;
+		alias = NULL;
 	}
 
-	object = dbus_g_proxy_new_for_name(connection, "org.bluez",
-						path, "org.bluez.Adapter");
-
-	dbus_g_proxy_call(object, "GetName", NULL, G_TYPE_INVALID,
-				G_TYPE_STRING, &adapter, G_TYPE_INVALID);
-
-	dbus_g_proxy_call(object, "GetRemoteName", NULL,
-				G_TYPE_STRING, address, G_TYPE_INVALID,
-				G_TYPE_STRING, &name, G_TYPE_INVALID);
-
-	object = dbus_g_proxy_new_for_name(connection, "org.bluez",
-						service, "org.bluez.Service");
-
-	dbus_g_proxy_call(object, "GetName", NULL, G_TYPE_INVALID,
-				G_TYPE_STRING, &profile, G_TYPE_INVALID);
-
-	if (name) {
-		if (g_strrstr(name, address))
-			device = g_strdup(name);
+	if (alias) {
+		if (g_strrstr(alias, address))
+			name = g_strdup(alias);
 		else
-			device = g_strdup_printf("%s (%s)", name, address);
+			name = g_strdup_printf("%s (%s)", alias, address);
 	} else
-		device = g_strdup(address);
+		name = g_strdup(address);
 
-	auth_dialog(path, address, service, uuid, device, profile, context);
+	text = g_strdup_printf("%d", passkey);
+	display_dialog(adapter, device, address, name, text, entered, context);
+	g_free(text);
 
-	line = g_strdup_printf(_("Authorization request for %s"), device);
-	g_free(device);
+	/* translators: this is a popup telling you a particular device
+	 * has asked for pairing */
+	line = g_strdup_printf(_("Pairing request for %s"), name);
 
-	show_notification(adapter ? adapter : _("Bluetooth device"),
+	g_free(name);
+
+	show_notification(_("Bluetooth device"),
+					line, _("Enter passkey"), 0,
+					G_CALLBACK(notification_closed));
+
+	g_free(line);
+
+	return TRUE;
+}
+
+static gboolean confirm_request(DBusGMethodInvocation *context,
+			DBusGProxy *device, guint passkey, gpointer user_data)
+{
+	DBusGProxy *adapter = user_data;
+	GHashTable *hash = NULL;
+	GValue *value;
+	const gchar *address, *alias;
+	gchar *name, *line, *text;
+
+	device_get_properties(device, &hash, NULL);
+
+	if (hash != NULL) {
+		value = g_hash_table_lookup(hash, "Address");
+		address = value ? g_value_get_string(value) : NULL;
+
+		value = g_hash_table_lookup(hash, "Name");
+		alias = value ? g_value_get_string(value) : NULL;
+	} else {
+		address = NULL;
+		alias = NULL;
+	}
+
+	if (alias) {
+		if (g_strrstr(alias, address))
+			name = g_strdup(alias);
+		else
+			name = g_strdup_printf("%s (%s)", alias, address);
+	} else
+		name = g_strdup(address);
+
+	text = g_strdup_printf("%d", passkey);
+	confirm_dialog(adapter, device, address, name, text, context);
+	g_free(text);
+
+	/* translators: this is a popup telling you a particular device
+	 * has asked for pairing */
+	line = g_strdup_printf(_("Confirmation request for %s"), name);
+
+	g_free(name);
+
+	show_notification(_("Bluetooth device"),
+					line, _("Confirm passkey"), 0,
+					G_CALLBACK(notification_closed));
+
+	g_free(line);
+
+	return TRUE;
+}
+
+static gboolean authorize_request(DBusGMethodInvocation *context,
+		DBusGProxy *device, const char *uuid, gpointer user_data)
+{
+	DBusGProxy *adapter = user_data;
+	GHashTable *hash = NULL;
+	GValue *value;
+	const gchar *address, *alias;
+	gchar *name, *line;
+
+	device_get_properties(device, &hash, NULL);
+
+	if (hash != NULL) {
+		value = g_hash_table_lookup(hash, "Address");
+		address = value ? g_value_get_string(value) : NULL;
+
+		value = g_hash_table_lookup(hash, "Name");
+		alias = value ? g_value_get_string(value) : NULL;
+	} else {
+		address = NULL;
+		alias = NULL;
+	}
+
+	if (alias) {
+		if (g_strrstr(alias, address))
+			name = g_strdup(alias);
+		else
+			name = g_strdup_printf("%s (%s)", alias, address);
+	} else
+		name = g_strdup(address);
+
+	auth_dialog(adapter, device, address, name, uuid, context);
+
+	line = g_strdup_printf(_("Authorization request for %s"), name);
+
+	g_free(name);
+
+	show_notification(_("Bluetooth device"),
 					line, _("Check authorization"), 0,
 					G_CALLBACK(notification_closed));
 
@@ -877,10 +787,10 @@ static gboolean auth_agent_authorize(PasskeyAgent *agent,
 	return TRUE;
 }
 
-static gboolean auth_agent_cancel(PasskeyAgent *agent,
-		const char *path, const char *address, const char *service,
-			const char *uuid, DBusGMethodInvocation *context)
+static gboolean cancel_request(DBusGMethodInvocation *context,
+							gpointer user_data)
 {
+	DBusGProxy *adapter = user_data;
 	GList *list;
 	GError *result;
 	struct input_data *input;
@@ -889,16 +799,10 @@ static gboolean auth_agent_cancel(PasskeyAgent *agent,
 	if (!input)
 		return FALSE;
 
-	input->path = g_strdup(path);
-	input->address = g_strdup(address);
-	input->service = g_strdup(service);
-	input->uuid = g_strdup(uuid);
+	input->path = g_strdup(dbus_g_proxy_get_path(adapter));
 
 	list = g_list_find_custom(input_list, input, input_compare);
 
-	g_free(input->uuid);
-	g_free(input->service);
-	g_free(input->address);
 	g_free(input->path);
 	g_free(input);
 
@@ -919,89 +823,59 @@ static gboolean auth_agent_cancel(PasskeyAgent *agent,
 	return TRUE;
 }
 
-static gboolean auth_agent_release(PasskeyAgent *agent, GError **error)
+static gboolean adapter_insert(GtkTreeModel *model, GtkTreePath *path,
+					GtkTreeIter *iter, gpointer user_data)
 {
-	registered_auth = 0;
+	BluetoothAgent *agent;
+	DBusGProxy *adapter;
 
-	return TRUE;
+	gtk_tree_model_get(model, iter, BLUETOOTH_COLUMN_PROXY, &adapter, -1);
+
+	if (adapter == NULL)
+		return FALSE;
+
+	agent = bluetooth_agent_new();
+
+	bluetooth_agent_set_pincode_func(agent, pincode_request, adapter);
+	bluetooth_agent_set_passkey_func(agent, passkey_request, adapter);
+	bluetooth_agent_set_display_func(agent, display_request, adapter);
+	bluetooth_agent_set_confirm_func(agent, confirm_request, adapter);
+	bluetooth_agent_set_authorize_func(agent, authorize_request, adapter);
+	bluetooth_agent_set_cancel_func(agent, cancel_request, adapter);
+
+	bluetooth_agent_register(agent, adapter);
+
+	return FALSE;
 }
 
-#include "auth-agent-glue.h"
-
-int register_agents(void)
+static void adapter_added(GtkTreeModel *model, GtkTreePath *path,
+					GtkTreeIter *iter, gpointer user_data)
 {
-	DBusGProxy *object;
-	GError *error = NULL;
-
-	if (registered_passkey && registered_auth)
-		return 0;
-
-	object = dbus_g_proxy_new_for_name(connection, "org.bluez",
-					"/org/bluez", "org.bluez.Security");
-
-	if (!registered_passkey) {
-
-		dbus_g_proxy_call(object, "RegisterDefaultPasskeyAgent",
-				&error, G_TYPE_STRING, PASSKEY_AGENT_PATH,
-					G_TYPE_INVALID, G_TYPE_INVALID);
-
-		if (error != NULL) {
-			g_error_free(error);
-			return -1;
-		}
-
-		registered_passkey = 1;
-	}
-
-	if (!registered_auth) {
-		dbus_g_proxy_call(object, "RegisterDefaultAuthorizationAgent",
-				&error,	G_TYPE_STRING, AUTH_AGENT_PATH,
-					G_TYPE_INVALID, G_TYPE_INVALID);
-
-		if (error != NULL) {
-			g_error_free(error);
-			return -1;
-		}
-
-		registered_auth = 1;
-	}
-
-	return 0;
+	adapter_insert(model, path, iter, user_data);
 }
 
-void unregister_agents(void)
+int setup_agents(void)
 {
-	registered_passkey = 0;
-	registered_auth = 0;
-}
-
-int setup_agents(DBusGConnection *conn)
-{
-	void *agent;
-
-	connection = dbus_g_connection_ref(conn);
-
-	dbus_g_object_type_install_info(PASSKEY_AGENT_OBJECT_TYPE,
-					&dbus_glib_passkey_agent_object_info);
-
-	dbus_g_object_type_install_info(AUTH_AGENT_OBJECT_TYPE,
-					&dbus_glib_auth_agent_object_info);
-
 	dbus_g_error_domain_register(AGENT_ERROR, "org.bluez.Error",
 							AGENT_ERROR_TYPE);
 
-	agent = passkey_agent_new(PASSKEY_AGENT_PATH);
+	client = bluetooth_client_new();
 
-	agent = auth_agent_new(AUTH_AGENT_PATH);
+	adapter_model = bluetooth_client_get_adapter_model(client);
+
+	g_signal_connect(G_OBJECT(adapter_model), "row-inserted",
+					G_CALLBACK(adapter_added), NULL);
+
+	gtk_tree_model_foreach(adapter_model, adapter_insert, NULL);
 
 	return 0;
 }
 
 void cleanup_agents(void)
 {
-	unregister_agents();
+	g_object_unref(adapter_model);
 
-	dbus_g_connection_unref(connection);
+	g_object_unref(client);
 }
 
 void show_agents(void)
@@ -1012,8 +886,3 @@ void show_agents(void)
 
 	disable_blinking();
 }
-
-void set_auto_authorize(gboolean value)
-{
-	auto_authorize = value;
-}
