@@ -27,11 +27,12 @@
 
 #include <gtk/gtk.h>
 
-#include "client.h"
+#include <dbus/dbus-glib.h>
+
+#include "bluetooth-client.h"
 
 static BluetoothClient *client;
-
-static GtkWidget *statusbar;
+static GtkTreeSelection *selection;
 
 static void delete_callback(GtkWidget *window, GdkEvent *event,
 						gpointer user_data)
@@ -41,114 +42,170 @@ static void delete_callback(GtkWidget *window, GdkEvent *event,
 	gtk_main_quit();
 }
 
-static void default_adapter_changed (GObject    *gobject,
-				     GParamSpec *arg1,
-				     gpointer    user_data)
+static void scan_callback(GtkWidget *button, gpointer user_data)
 {
-	char *adapter;
+	bluetooth_client_start_discovery(client);
+}
 
-	g_object_get (gobject, "default-adapter", &adapter, NULL);
-	g_message ("default adapter changed: %s", adapter);
-	g_free (adapter);
+static void select_callback(GtkTreeSelection *selection, gpointer user_data)
+{
+	GtkTreeModel *model;
+	GtkTreeIter iter;
+	gboolean selected;
+
+	selected = gtk_tree_selection_get_selected(selection, &model, &iter);
+}
+
+static void row_inserted(GtkTreeModel *model, GtkTreePath *path,
+				GtkTreeIter *iter, gpointer user_data)
+{
+	GtkTreeView *tree = user_data;
+
+	gtk_tree_view_expand_all(tree);
+}
+
+static void proxy_to_text(GtkTreeViewColumn *column, GtkCellRenderer *cell,
+		GtkTreeModel *model, GtkTreeIter *iter, gpointer user_data)
+{
+	DBusGProxy *proxy;
+	gchar *path;
+
+	gtk_tree_model_get(model, iter, BLUETOOTH_COLUMN_PROXY, &proxy, -1);
+
+	if (proxy == NULL) {
+		g_object_set(cell, "text", "", NULL);
+		return;
+	}
+
+	path = g_path_get_basename(dbus_g_proxy_get_path(proxy));
+
+	g_object_set(cell, "text", path, NULL);
+
+	g_free(path);
+
+	g_object_unref(proxy);
+}
+
+static void type_to_text(GtkTreeViewColumn *column, GtkCellRenderer *cell,
+		GtkTreeModel *model, GtkTreeIter *iter, gpointer user_data)
+{
+	guint type;
+
+	gtk_tree_model_get(model, iter, BLUETOOTH_COLUMN_TYPE, &type, -1);
+
+	g_object_set(cell, "text", bluetooth_type_to_string(type), NULL);
 }
 
 static void create_window(void)
 {
 	GtkWidget *window;
 	GtkWidget *vbox;
+	GtkWidget *toolbar;
+	GtkToolItem *item;
 	GtkWidget *scrolled;
 	GtkWidget *tree;
 	GtkTreeModel *model;
+	GtkTreeModel *sorted;
+	GtkWidget *statusbar;
 
 	window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
-
 	gtk_window_set_title(GTK_WINDOW(window), "Test client");
-
 	gtk_window_set_icon_name(GTK_WINDOW(window), "bluetooth");
-
 	gtk_window_set_position(GTK_WINDOW(window), GTK_WIN_POS_CENTER);
+	gtk_window_set_default_size(GTK_WINDOW(window), 800, 600);
+	g_signal_connect(G_OBJECT(window), "delete-event",
+					G_CALLBACK(delete_callback), NULL);
 
-	gtk_window_set_default_size(GTK_WINDOW(window), 920, 500);
+	vbox = gtk_vbox_new(FALSE, 0);
+	gtk_container_add(GTK_CONTAINER(window), vbox);
 
-	vbox = gtk_vbox_new(FALSE, 6);
+	toolbar = gtk_toolbar_new();
+	gtk_toolbar_set_show_arrow(GTK_TOOLBAR(toolbar), FALSE);
+	gtk_box_pack_start(GTK_BOX(vbox), toolbar, FALSE, TRUE, 0);
+
+	item = gtk_tool_button_new_from_stock(GTK_STOCK_REFRESH);
+	gtk_toolbar_insert(GTK_TOOLBAR(toolbar), item, -1);
+	g_signal_connect(item, "clicked", G_CALLBACK(scan_callback), NULL);
 
 	scrolled = gtk_scrolled_window_new(NULL, NULL);
-
 	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrolled),
 				GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
-
 	gtk_container_add(GTK_CONTAINER(vbox), scrolled);
 
 	tree = gtk_tree_view_new();
-
 	gtk_tree_view_set_headers_visible(GTK_TREE_VIEW(tree), TRUE);
-
 	gtk_tree_view_set_rules_hint(GTK_TREE_VIEW(tree), TRUE);
+	gtk_widget_grab_focus(GTK_WIDGET(tree));
+	gtk_container_add(GTK_CONTAINER(scrolled), tree);
 
-	model = bluetooth_client_get_model(client);
-	//model = bluetooth_client_get_model_adapter_list(client);
-
-	gtk_tree_view_set_model(GTK_TREE_VIEW(tree), model);
-
-	g_object_unref(model);
-
-	gtk_tree_view_expand_all(GTK_TREE_VIEW(tree));
-
-	gtk_tree_view_insert_column_with_attributes(GTK_TREE_VIEW(tree), -1,
-					"Path", gtk_cell_renderer_text_new(),
-						"text", COLUMN_PATH, NULL);
-
-	gtk_tree_view_insert_column_with_attributes(GTK_TREE_VIEW(tree), -1,
-					"Active", gtk_cell_renderer_text_new(),
-						"text", COLUMN_ACTIVE, NULL);
+	gtk_tree_view_insert_column_with_data_func(GTK_TREE_VIEW(tree), -1,
+					"Proxy", gtk_cell_renderer_text_new(),
+						proxy_to_text, NULL, NULL);
 
 	gtk_tree_view_insert_column_with_attributes(GTK_TREE_VIEW(tree), -1,
 					"Address", gtk_cell_renderer_text_new(),
-						"text", COLUMN_ADDRESS, NULL);
+					"text", BLUETOOTH_COLUMN_ADDRESS, NULL);
 
 	gtk_tree_view_insert_column_with_attributes(GTK_TREE_VIEW(tree), -1,
-					"Class", gtk_cell_renderer_text_new(),
-						"text", COLUMN_CLASS, NULL);
-
-	gtk_tree_view_insert_column_with_attributes(GTK_TREE_VIEW(tree), -1,
-					"RSSI", gtk_cell_renderer_text_new(),
-						"text", COLUMN_RSSI, NULL);
+					"Alias", gtk_cell_renderer_text_new(),
+					"text", BLUETOOTH_COLUMN_ALIAS, NULL);
 
 	gtk_tree_view_insert_column_with_attributes(GTK_TREE_VIEW(tree), -1,
 					"Name", gtk_cell_renderer_text_new(),
-						"text", COLUMN_NAME, NULL);
+					"text", BLUETOOTH_COLUMN_NAME, NULL);
 
-	gtk_tree_view_insert_column_with_attributes(GTK_TREE_VIEW(tree), -1,
+	gtk_tree_view_insert_column_with_data_func(GTK_TREE_VIEW(tree), -1,
 					"Type", gtk_cell_renderer_text_new(),
-						"text", COLUMN_TYPE, NULL);
+						type_to_text, NULL, NULL);
 
 	gtk_tree_view_insert_column_with_attributes(GTK_TREE_VIEW(tree), -1,
-					"Bonded", gtk_cell_renderer_text_new(),
-						"text", COLUMN_BONDED, NULL);
+					"Icon", gtk_cell_renderer_text_new(),
+					"text", BLUETOOTH_COLUMN_ICON, NULL);
+
+	gtk_tree_view_insert_column_with_attributes(GTK_TREE_VIEW(tree), -1,
+					"RSSI", gtk_cell_renderer_text_new(),
+					"text", BLUETOOTH_COLUMN_RSSI, NULL);
+
+	gtk_tree_view_insert_column_with_attributes(GTK_TREE_VIEW(tree), -1,
+					"Default", gtk_cell_renderer_text_new(),
+					"text", BLUETOOTH_COLUMN_DEFAULT, NULL);
+
+	gtk_tree_view_insert_column_with_attributes(GTK_TREE_VIEW(tree), -1,
+					"Paired", gtk_cell_renderer_text_new(),
+					"text", BLUETOOTH_COLUMN_PAIRED, NULL);
 
 	gtk_tree_view_insert_column_with_attributes(GTK_TREE_VIEW(tree), -1,
 					"Trusted", gtk_cell_renderer_text_new(),
-						"text", COLUMN_TRUSTED, NULL);
+					"text", BLUETOOTH_COLUMN_TRUSTED, NULL);
 
 	gtk_tree_view_insert_column_with_attributes(GTK_TREE_VIEW(tree), -1,
 					"Connected", gtk_cell_renderer_text_new(),
-						"text", COLUMN_CONNECTED, NULL);
+					"text", BLUETOOTH_COLUMN_CONNECTED, NULL);
 
-	gtk_widget_grab_focus(GTK_WIDGET(tree));
+	gtk_tree_view_insert_column_with_attributes(GTK_TREE_VIEW(tree), -1,
+					"Discovering", gtk_cell_renderer_text_new(),
+					"text", BLUETOOTH_COLUMN_DISCOVERING, NULL);
 
-	gtk_container_add(GTK_CONTAINER(scrolled), tree);
+	model = bluetooth_client_get_model(client);
+	sorted = gtk_tree_model_sort_new_with_model(model);
+	gtk_tree_sortable_set_sort_column_id(GTK_TREE_SORTABLE(sorted),
+				BLUETOOTH_COLUMN_RSSI, GTK_SORT_DESCENDING);
+	gtk_tree_view_set_model(GTK_TREE_VIEW(tree), sorted);
+	g_signal_connect(G_OBJECT(model), "row-inserted",
+					G_CALLBACK(row_inserted), tree);
+	g_object_unref(sorted);
+	g_object_unref(model);
 
-	gtk_container_add(GTK_CONTAINER(window), vbox);
+	selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(tree));
+	gtk_tree_selection_set_mode(selection, GTK_SELECTION_SINGLE);
+	g_signal_connect(G_OBJECT(selection), "changed",
+				G_CALLBACK(select_callback), NULL);
+
+	gtk_tree_view_expand_all(GTK_TREE_VIEW(tree));
 
 	statusbar = gtk_statusbar_new();
-
 	gtk_box_pack_start(GTK_BOX(vbox), statusbar, FALSE, FALSE, 0);
 
-	g_signal_connect(G_OBJECT(window), "delete-event",
-					G_CALLBACK(delete_callback), NULL);
-	g_signal_connect(G_OBJECT(client), "notify::default-adapter",
-			 G_CALLBACK(default_adapter_changed), client);
-
 	gtk_widget_show_all(window);
 }
 
