@@ -21,12 +21,5 @@
  *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
  *
  */
- 
-typedef void (*DialogConnectCallbackFunc) (const gchar *address, gpointer data);
-typedef void (*DialogCleanupCallbackFunc) (gpointer data);
 
 gboolean show_confirm_dialog(void);
-gboolean show_remove_dialog(void);
-void show_device_dialog(DialogConnectCallbackFunc callback,
-				DialogCleanupCallbackFunc cleanup,
-				GtkTreeModelFilterVisibleFunc filter);
