@@ -26,13 +26,14 @@
 #include <config.h>
 #endif
 
-#include <gconf/gconf-client.h>
-
 #include <glib/gi18n.h>
-
 #include <gtk/gtk.h>
 
-#include "client.h"
+#include <gconf/gconf-client.h>
+
+#include <bluetooth-client.h>
+
+#include "killswitch.h"
 #include "general.h"
 
 typedef enum {
@@ -44,11 +45,11 @@ typedef enum {
 static int icon_policy = ICON_POLICY_PRESENT;
 
 #define PREF_DIR		"/apps/bluetooth-manager"
-#define PREF_USE_HAL		PREF_DIR "/use_hal"
 #define PREF_ICON_POLICY	PREF_DIR "/icon_policy"
-#define PREF_AUTO_AUTHORIZE	PREF_DIR "/auto_authorize"
+#if 0
 #define PREF_RECEIVE_ENABLED	PREF_DIR "/receive_enabled"
 #define PREF_SHARING_ENABLED	PREF_DIR "/sharing_enabled"
+#endif
 
 static GConfEnumStringPair icon_policy_enum_map [] = {
 	{ ICON_POLICY_NEVER,	"never"		},
@@ -59,11 +60,9 @@ static GConfEnumStringPair icon_policy_enum_map [] = {
 
 static GConfClient* gconf;
 
+#if 0
 static GtkWidget *button_receive;
 static GtkWidget *button_sharing;
-static GtkWidget *button_autoauth;
-#ifdef HAVE_HAL
-static GtkWidget *button_usehal;
 #endif
 static GtkWidget *button_never;
 static GtkWidget *button_always;
@@ -114,26 +113,7 @@ static void policy_callback(GtkWidget *button, gpointer user_data)
 		update_icon_policy(button);
 }
 
-#ifdef HAVE_HAL
-static void usehal_callback(GtkWidget *button, gpointer user_data)
-{
-	gboolean value;
-
-	value = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(button));
-
-	gconf_client_set_bool(gconf, PREF_USE_HAL, value, NULL);
-}
-#endif
-
-static void autoauth_callback(GtkWidget *button, gpointer user_data)
-{
-	gboolean value;
-
-	value = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(button));
-
-	gconf_client_set_bool(gconf, PREF_AUTO_AUTHORIZE, value, NULL);
-}
-
+#if 0
 static void receive_callback(GtkWidget *button, gpointer user_data)
 {
 	gboolean value;
@@ -151,6 +131,7 @@ static void sharing_callback(GtkWidget *button, gpointer user_data)
 
 	gconf_client_set_bool(gconf, PREF_SHARING_ENABLED, value, NULL);
 }
+#endif
 
 GtkWidget *create_label(const gchar *str)
 {
@@ -178,11 +159,14 @@ GtkWidget *create_general(void)
 	GtkWidget *vbox;
 	GtkWidget *label;
 	GSList *group = NULL;
+#if 0
 	gboolean value;
+#endif
 
 	mainbox = gtk_vbox_new(FALSE, 24);
 	gtk_container_set_border_width(GTK_CONTAINER(mainbox), 12);
 
+#if 0
 	vbox = gtk_vbox_new(FALSE, 6);
 	gtk_box_pack_start(GTK_BOX(mainbox), vbox, FALSE, FALSE, 0);
 
@@ -204,40 +188,6 @@ GtkWidget *create_general(void)
 	gtk_box_pack_start(GTK_BOX(vbox), button_sharing, FALSE, FALSE, 0);
 	g_signal_connect(G_OBJECT(button_sharing), "toggled",
 					G_CALLBACK(sharing_callback), NULL);
-
-	vbox = gtk_vbox_new(FALSE, 6);
-	gtk_box_pack_start(GTK_BOX(mainbox), vbox, FALSE, FALSE, 0);
-
-	label = create_label(_("Authorization requests"));
-	gtk_box_pack_start(GTK_BOX(vbox), label, FALSE, FALSE, 0);
-
-	button_autoauth = gtk_check_button_new_with_label(
-			_("Automatically authorize incoming requests"));
-	value = gconf_client_get_bool(gconf, PREF_AUTO_AUTHORIZE, NULL);
-	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(button_autoauth), value);
-	gtk_box_pack_start(GTK_BOX(vbox), button_autoauth, FALSE, FALSE, 0);
-	g_signal_connect(G_OBJECT(button_autoauth), "toggled",
-					G_CALLBACK(autoauth_callback), NULL);
-
-#ifdef HAVE_HAL
-	vbox = gtk_vbox_new(FALSE, 6);
-	gtk_box_pack_start(GTK_BOX(mainbox), vbox, FALSE, FALSE, 0);
-
-	label = create_label(_("Hardware database"));
-	gtk_box_pack_start(GTK_BOX(vbox), label, FALSE, FALSE, 0);
-
-	button_usehal = gtk_check_button_new_with_label(
-				_("Select class of device automatically"));
-	value = gconf_client_get_bool(gconf, PREF_USE_HAL, NULL);
-	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(button_usehal), value);
-	gtk_box_pack_start(GTK_BOX(vbox), button_usehal, FALSE, FALSE, 0);
-	g_signal_connect(G_OBJECT(button_usehal), "toggled",
-					G_CALLBACK(usehal_callback), NULL);
-
-	if (value == TRUE)
-		hide_class_widget();
-	else
-		show_class_widget();
 #endif
 
 	vbox = gtk_vbox_new(FALSE, 6);
@@ -269,6 +219,9 @@ GtkWidget *create_general(void)
 
 	update_icon_policy(NULL);
 
+	vbox = create_killswitch();
+	gtk_box_pack_start(GTK_BOX(mainbox), vbox, FALSE, FALSE, 0);
+
 	return mainbox;
 }
 
@@ -281,24 +234,6 @@ static void gconf_callback(GConfClient *client, guint cnxn_id,
 	if (value == NULL)
 		return;
 
-#ifdef HAVE_HAL
-	if (g_str_equal(entry->key, PREF_USE_HAL) == TRUE) {
-		gboolean use_hal;
-
-		use_hal = gconf_value_get_bool(value);
-		if (use_hal == TRUE)
-			hide_class_widget();
-		else
-			show_class_widget();
-
-		if (!button_usehal)
-			return;
-
-		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(button_usehal),
-								use_hal);
-	}
-#endif
-
 	if (g_str_equal(entry->key, PREF_ICON_POLICY) == TRUE) {
 		const char *str;
 
@@ -310,14 +245,7 @@ static void gconf_callback(GConfClient *client, guint cnxn_id,
 		update_icon_policy(NULL);
 	}
 
-	if (g_str_equal(entry->key, PREF_AUTO_AUTHORIZE) == TRUE) {
-		if (!button_autoauth)
-			return;
-
-		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(button_autoauth),
-						gconf_value_get_bool(value));
-	}
-
+#if 0
 	if (g_str_equal(entry->key, PREF_RECEIVE_ENABLED) == TRUE) {
 		if (!button_receive)
 			return;
@@ -333,6 +261,7 @@ static void gconf_callback(GConfClient *client, guint cnxn_id,
 		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(button_sharing),
 						gconf_value_get_bool(value));
 	}
+#endif
 }
 
 void setup_general(void)
@@ -351,9 +280,13 @@ void setup_general(void)
 
 	gconf_client_notify_add(gconf, PREF_DIR,
 					gconf_callback, NULL, NULL, NULL);
+
+	setup_killswitch();
 }
 
 void cleanup_general(void)
 {
+	cleanup_killswitch();
+
 	g_object_unref(gconf);
 }
