@@ -26,6 +26,3 @@ GtkWidget *create_label(const gchar *str);
 GtkWidget *create_general(void);
 void setup_general(void);
 void cleanup_general(void);
-
-void hide_class_widget(void);
-void show_class_widget(void);
