@@ -27,16 +27,20 @@
 
 #include <sys/time.h>
 
-#include <dbus/dbus-glib.h>
-
 #include <glib/gi18n.h>
 #include <glib/gstdio.h>
-
+#include <gio/gio.h>
 #include <gtk/gtk.h>
 
+#include <dbus/dbus-glib.h>
+#include <dbus/dbus-glib-lowlevel.h>
+
+#include <obex-agent.h>
+
+#include "helper.h"
 #include "marshal.h"
 
-#include "dialog.h"
+#define AGENT_PATH "/org/bluez/agent/sendto"
 
 static DBusGConnection *conn = NULL;
 
@@ -60,6 +64,18 @@ static int file_index = 0;
 static gint64 first_update = 0;
 static gint64 last_update = 0;
 
+static gchar *filename_to_path(const gchar *filename)
+{
+	GFile *file;
+	gchar *path;
+
+	file = g_file_new_for_commandline_arg(filename);
+	path = g_file_get_path(file);
+	g_object_unref(file);
+
+	return path;
+}
+
 static gint64 get_system_time(void)
 {
 	struct timeval tmp;
@@ -70,6 +86,44 @@ static gint64 get_system_time(void)
 			(gint64) tmp.tv_sec * G_GINT64_CONSTANT(1000000);
 }
 
+static gchar *format_time(gint seconds)
+{
+	gint hours, minutes;
+
+	if (seconds < 0)
+		seconds = 0;
+
+	if (seconds < 60)
+		return g_strdup_printf(ngettext("%'d second",
+					"%'d seconds", seconds), seconds);
+
+	if (seconds < 60 * 60) {
+		minutes = (seconds + 30) / 60;
+		return g_strdup_printf(ngettext("%'d minute",
+					"%'d minutes", minutes), minutes);
+	}
+
+	hours = seconds / (60 * 60);
+
+	if (seconds < 60 * 60 * 4) {
+		gchar *res, *h, *m;
+
+		minutes = (seconds - hours * 60 * 60 + 30) / 60;
+
+		h = g_strdup_printf(ngettext("%'d hour",
+					"%'d hours", hours), hours);
+		m = g_strdup_printf(ngettext("%'d minute",
+					"%'d minutes", minutes), minutes);
+		res = g_strconcat(h, ", ", m, NULL);
+		g_free(h);
+		g_free(m);
+		return res;
+	}
+
+	return g_strdup_printf(ngettext("approximately %'d hour",
+				"approximately %'d hours", hours), hours);
+}
+
 static void response_callback(GtkWidget *dialog,
 					gint response, gpointer user_data)
 {
@@ -266,14 +320,7 @@ static void transfer_progress(DBusGProxy *proxy,
 
 	remaining_time = (total_size - current_sent) / transfer_rate;
 
-	if (remaining_time >= 3600)
-		time = g_strdup_printf(_("%d:%02d:%02d Remaining"),
-					remaining_time / 3600,
-					(remaining_time % 3600) / 60,
-					(remaining_time % 3600) % 60);
-	else
-		time = g_strdup_printf(_("%d:%02d Remaining"),
-				remaining_time / 60, remaining_time % 60);
+	time = format_time(remaining_time);
 
 	if (transfer_rate >= 3000)
 		rate = g_strdup_printf(_("%d KB/s"), transfer_rate / 1000);
@@ -416,57 +463,206 @@ static void create_notify(DBusGProxy *proxy,
 							G_TYPE_INVALID);
 }
 
+static gchar *get_name(DBusGProxy *device)
+{
+	GHashTable *hash;
+
+	if (dbus_g_proxy_call(device, "GetProperties", NULL,
+			G_TYPE_INVALID, dbus_g_type_get_map("GHashTable",
+						G_TYPE_STRING, G_TYPE_VALUE),
+					&hash, G_TYPE_INVALID) != FALSE) {
+		GValue *value;
+		gchar *name;
+
+		value = g_hash_table_lookup(hash, "Name");
+		name = value ? g_value_dup_string(value) : NULL;
+		g_hash_table_destroy(hash);
+
+		return name;
+	}
+
+	return NULL;
+}
+
 static gchar *get_device_name(const gchar *address)
 {
 	DBusGConnection *connection;
-	DBusGProxy *manager, *proxy;
-	gchar *adapter, *name;
+	DBusGProxy *manager;
+	GPtrArray *adapters;
+	gchar *name;
+	guint i;
+
+	name = NULL;
 
 	connection = dbus_g_bus_get(DBUS_BUS_SYSTEM, NULL);
 	if (connection == NULL)
-		return NULL;
+		return name;
 
 	manager = dbus_g_proxy_new_for_name(connection, "org.bluez",
-					"/org/bluez", "org.bluez.Manager");
+						"/", "org.bluez.Manager");
 	if (manager == NULL) {
 		dbus_g_connection_unref(connection);
-		return NULL;
+		return name;
 	}
 
-	if (dbus_g_proxy_call(manager, "DefaultAdapter", NULL, G_TYPE_INVALID,
-			G_TYPE_STRING, &adapter, G_TYPE_INVALID) == FALSE) {
+	if (dbus_g_proxy_call(manager, "ListAdapters", NULL,
+			G_TYPE_INVALID, dbus_g_type_get_collection("GPtrArray",
+						DBUS_TYPE_G_OBJECT_PATH),
+					&adapters, G_TYPE_INVALID) == FALSE) {
 		g_object_unref(manager);
 		dbus_g_connection_unref(connection);
-		return NULL;
+		return name;
 	}
 
+	for (i = 0; i < adapters->len && name == NULL; i++) {
+		DBusGProxy *adapter;
+		char *device_path;
+
+		adapter = dbus_g_proxy_new_for_name(connection, "org.bluez",
+						g_ptr_array_index(adapters, i),
+							"org.bluez.Adapter");
+
+		if (dbus_g_proxy_call(adapter, "FindDevice", NULL,
+					G_TYPE_STRING, address, G_TYPE_INVALID,
+					DBUS_TYPE_G_OBJECT_PATH, &device_path,
+						G_TYPE_INVALID) == TRUE) {
+			DBusGProxy *device;
+			device = dbus_g_proxy_new_for_name(connection,
+						"org.bluez", device_path,
+							"org.bluez.Device");
+			name = get_name(device);
+			g_object_unref(device);
+			break;
+		}
+
+		g_object_unref(adapter);
+	}
+
+	g_ptr_array_free(adapters, TRUE);
 	g_object_unref(manager);
 
-	proxy = dbus_g_proxy_new_for_name(connection, "org.bluez",
-						adapter, "org.bluez.Adapter");
-	if (proxy == NULL) {
-		g_free(adapter);
-		dbus_g_connection_unref(connection);
-		return NULL;
-	}
-
-	g_free(adapter);
-
-	if (dbus_g_proxy_call(proxy, "GetRemoteName", NULL,
-			G_TYPE_STRING, address, G_TYPE_INVALID,
-			G_TYPE_STRING, &name, G_TYPE_INVALID) == FALSE) {
-		g_object_unref(proxy);
-		dbus_g_connection_unref(connection);
-		return NULL;
-	}
-
-	g_object_unref(proxy);
-
 	dbus_g_connection_unref(connection);
 
 	return name;
 }
 
+static gboolean request_callback(DBusGMethodInvocation *context,
+				DBusGProxy *transfer, gpointer user_data)
+{
+	gchar *filename = option_files[file_index];
+	gchar *basename, *text, *markup;
+	GHashTable *hash;
+
+	if (dbus_g_proxy_call(transfer,
+				"GetProperties", NULL, G_TYPE_INVALID,
+				dbus_g_type_get_map("GHashTable",
+						G_TYPE_STRING, G_TYPE_VALUE),
+					&hash, G_TYPE_INVALID) == TRUE) {
+		GValue *value;
+
+		value = g_hash_table_lookup(hash, "Size");
+		if (value) {
+			current_size = g_value_get_uint64(value);
+
+			last_update = get_system_time();
+		}
+
+		g_hash_table_destroy(hash);
+	}
+
+	text = g_path_get_dirname(filename);
+	gtk_label_set_markup(GTK_LABEL(label_from), text);
+	g_free(text);
+
+	basename = g_path_get_basename(filename);
+	text = g_strdup_printf(_("Sending %s"), basename);
+	g_free(basename);
+	markup = g_markup_printf_escaped("<i>%s</i>", text);
+	gtk_label_set_markup(GTK_LABEL(label_status), markup);
+	g_free(markup);
+	g_free(text);
+
+	text = g_strdup_printf(_("Sending file %d of %d"),
+						file_index + 1, file_count);
+	gtk_progress_bar_set_text(GTK_PROGRESS_BAR(progress), text);
+	g_free(text);
+
+	dbus_g_method_return(context, "");
+
+	return TRUE;
+}
+
+static gboolean progress_callback(DBusGMethodInvocation *context,
+				DBusGProxy *transfer, guint64 transferred,
+							gpointer user_data)
+{
+	transfer_progress(NULL, transferred, NULL);
+
+	dbus_g_method_return(context);
+
+	return TRUE;
+}
+
+static gboolean complete_callback(DBusGMethodInvocation *context,
+				DBusGProxy *transfer, gpointer user_data)
+{
+	total_sent += current_size;
+
+	file_index++;
+
+	dbus_g_method_return(context);
+
+	return TRUE;
+}
+
+static gboolean release_callback(DBusGMethodInvocation *context,
+							gpointer user_data)
+{
+	dbus_g_method_return(context);
+
+	gtk_label_set_markup(GTK_LABEL(label_status), NULL);
+
+	gtk_widget_destroy(dialog);
+
+	gtk_main_quit();
+
+	return TRUE;
+}
+
+static void send_notify(DBusGProxy *proxy,
+				DBusGProxyCall *call, void *user_data)
+{
+	GError *error = NULL;
+
+	if (dbus_g_proxy_end_call(proxy, call, &error,
+						G_TYPE_INVALID) == FALSE) {
+		gchar *text, *message;
+
+		message = get_error_message(error);
+
+		text = g_strdup_printf("<span foreground=\"red\">%s</span>",
+								message);
+		gtk_label_set_markup(GTK_LABEL(label_status), text);
+		g_free(text);
+
+		g_free(message);
+
+		gtk_dialog_set_response_sensitive(GTK_DIALOG(dialog),
+						GTK_RESPONSE_CLOSE, TRUE);
+		return;
+	}
+
+	gtk_progress_bar_set_text(GTK_PROGRESS_BAR(progress), NULL);
+
+	first_update = get_system_time();
+}
+
+static void value_free(GValue *value)
+{
+	g_value_unset(value);
+	g_free(value);
+}
+
 static void name_owner_changed(DBusGProxy *proxy, const char *name,
 			const char *prev, const char *new, gpointer user_data)
 {
@@ -526,17 +722,10 @@ int main(int argc, char *argv[])
 	file_count = g_strv_length(option_files);
 
 	for (i = 0; i < file_count; i++) {
-		gchar *filename = NULL;
+		gchar *filename;
 		struct stat st;
 
-		if (g_str_has_prefix(option_files[i], "file://") == TRUE) {
-			filename = g_filename_from_uri(option_files[i],
-								NULL, NULL);
-		} else if (option_files[i][0] != '/') {
-			gchar *dir = g_get_current_dir();
-			filename = g_build_filename(dir, option_files[i], NULL);
-			g_free(dir);
-		}
+		filename = filename_to_path(option_files[i]);
 
 		if (filename != NULL) {
 			g_free(option_files[i]);
@@ -584,19 +773,54 @@ int main(int argc, char *argv[])
 
 	create_window();
 
-	proxy = dbus_g_proxy_new_for_name(conn, "org.openobex",
+	proxy = dbus_g_proxy_new_for_name_owner(conn, "org.openobex.client",
+					"/", "org.openobex.Client", NULL);
+
+	if (proxy == NULL) {
+		proxy = dbus_g_proxy_new_for_name(conn, "org.openobex",
 				"/org/openobex", "org.openobex.Manager");
 
-	dbus_g_proxy_add_signal(proxy, "NameOwnerChanged",
-		G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_INVALID);
+		dbus_g_proxy_add_signal(proxy, "NameOwnerChanged",
+					G_TYPE_STRING, G_TYPE_STRING,
+					G_TYPE_STRING, G_TYPE_INVALID);
 
-	dbus_g_proxy_connect_signal(proxy, "NameOwnerChanged",
+		dbus_g_proxy_connect_signal(proxy, "NameOwnerChanged",
 				G_CALLBACK(name_owner_changed), NULL, NULL);
 
-	dbus_g_proxy_begin_call(proxy, "CreateBluetoothSession",
-				create_notify, NULL, NULL,
-				G_TYPE_STRING, option_device,
-				G_TYPE_STRING, "opp", G_TYPE_INVALID);
+		dbus_g_proxy_begin_call(proxy, "CreateBluetoothSession",
+					create_notify, NULL, NULL,
+					G_TYPE_STRING, option_device,
+					G_TYPE_STRING, "opp", G_TYPE_INVALID);
+	} else {
+		GHashTable *hash = NULL;
+		GValue *value;
+		ObexAgent *agent;
+
+		agent = obex_agent_new();
+
+		obex_agent_set_release_func(agent, release_callback, NULL);
+		obex_agent_set_request_func(agent, request_callback, NULL);
+		obex_agent_set_progress_func(agent, progress_callback, NULL);
+		obex_agent_set_complete_func(agent, complete_callback, NULL);
+
+		obex_agent_setup(agent, AGENT_PATH);
+
+		hash = g_hash_table_new_full(g_str_hash, g_str_equal,
+					g_free, (GDestroyNotify) value_free);
+
+		value = g_new0(GValue, 1);
+		g_value_init(value, G_TYPE_STRING);
+		g_value_set_string(value, option_device);
+		g_hash_table_insert(hash, g_strdup("Destination"), value);
+
+		dbus_g_proxy_begin_call(proxy, "SendFiles",
+				send_notify, NULL, NULL,
+				dbus_g_type_get_map("GHashTable",
+					G_TYPE_STRING, G_TYPE_VALUE), hash,
+				G_TYPE_STRV, option_files,
+				DBUS_TYPE_G_OBJECT_PATH, AGENT_PATH,
+							G_TYPE_INVALID);
+	}
 
 	gtk_main();
 
